from django import forms
from django.db import models

import django_filters


class FilterMixin(object):
    def get_filterset(self):
        """
        Returns an instance of the filterset to be used in this view.
        """
        self.filterset = self.filterset_class(**self.filterset_kwargs)

        # filter out choices which have no result to avoid filter pollution
        # with choices which would empty out results
        for name, field in self.filterset.form.fields.items():
            try:
                mf = self.model._meta.get_field(name)
            except Exception:
                continue

            if not isinstance(mf, models.ForeignKey):
                continue

            field.queryset = field.queryset.annotate(
                c=models.Count(mf.related_query_name())
            ).filter(c__gt=0)

    def get_filterset_data(self):
        return self.request.GET.copy()

    def get_filterset_kwargs(self):
        """
        Returns the keyword arguments for instanciating the filterset.
        """
        return {
            'data': self.filterset_data,
            'request': self.request,
            'queryset': self.queryset,
        }

    def get_filterset_meta_filter_overrides(self):
        return {
            models.CharField: {
               'filterset_class': django_filters.CharFilter,
               'extra': lambda f: {
                   'lookup_expr': 'icontains',
               },
            },
        }

    def get_filter_fields(self):
        return []

    def get_filterset_form_class(self):
        return type(
            'FiltersetForm',
            (forms.Form,),
            {
                '_layout': self.filterset_form_layout,
            }
        )

    def get_filterset_form_layout(self):
        return None

    def get_filterset_meta_attributes(self):
        return dict(
            model=self.model,
            fields=self.filter_fields,
            filter_overrides=self.filterset_meta_filter_overrides,
            form=self.filterset_form_class,
        )

    def get_filterset_meta_class(self):
        return type('Meta', (object,), self.filterset_meta_attributes)

    def get_filterset_extra_class_attributes(self):
        return dict()

    def get_filterset_class_attributes(self):
        res = dict(Meta=self.filterset_meta_class)
        res.update(self.filterset_extra_class_attributes)
        return res

    def get_filterset_class(self):
        return type(
            '{}FilterSet'.format(self.model.__name__),
            (django_filters.FilterSet,),
            self.filterset_class_attributes
        )



from crudlfap import shortcuts as crudlfap

from .models import Song


class SongMixin:
    allowed_groups = 'any'

    def get_exclude(self):
        if not self.request.user.is_staff:
            return ['owner']
        return super().get_exclude()


class SongCreateView(SongMixin, crudlfap.CreateView):
    def form_valid(self):
        self.form.instance.owner = self.request.user
        return super().form_valid()


class SongRouter(crudlfap.Router):
    fields = '__all__'
    icon = 'music'
    model = Song

    views = [
        crudlfap.DeleteView.clone(SongMixin),
        crudlfap.UpdateView.clone(SongMixin),
        SongCreateView,
        crudlfap.DetailView.clone(
            authenticate=False,
        ),
        crudlfap.ListView.clone(
            authenticate=False,
            filter_fields=['artist'],
            search_fields=['artist__name', 'name'],
        ),
    ]

    def get_queryset(self, view):
        user = view.request.user

        if user.is_staff or user.is_superuser:
            return self.model.objects.all()
        elif not user.is_authenticated:
            return self.model.objects.none()

        return self.model.objects.filter(owner=user)


SongRouter().register()



from crudlfap.models import Controller, URL
from crudlfap.router import Router
from crudlfap.views import generic

from django import forms
from django.contrib.auth.models import Group, Permission


class ControllerRouter(Router):
    model = Controller

    views = [
        generic.DetailView,
        generic.ListView.clone(
            search_fields=(
                'app',
                'model',
            ),
            table_fields=(
                'app',
                'model',
            ),
        ),
    ]


# useless ?
# ControllerRouter().register()


class AuthorizeView(generic.ObjectFormView):
    material_icon = 'lock'

    class form_class(forms.Form):
        groups = forms.ModelMultipleChoiceField(
            queryset=Group.objects.all(),
            required=False,
            widget=forms.CheckboxSelectMultiple,
        )

    def get_initial(self):
        kwargs = dict(
            codename=self.object.codename,
        )

        if self.object.model:
            kwargs['content_type'] = self.object.content_type

        perm = Permission.objects.filter(**kwargs).first()

        if not perm:
            return dict()

        return dict(
            groups=perm.group_set.all(),
        )

    def form_valid(self):
        perm = self.object.get_or_create_permission()
        for group in self.form.cleaned_data['groups']:
            group.permissions.add(perm)
        return super().form_valid()


class URLRouter(Router):
    model = URL
    material_icon = 'link'

    views = [
        generic.DetailView,
        AuthorizeView,
        generic.ListView.clone(
            search_fields=(
                'name',
                'controller__app',
                'controller__model',
            ),
            table_fields=(
                'controller',
                'id',
                'fullurlpath',
            ),
        ),
    ]


URLRouter().register()



from crudlfap import shortcuts as crudlfap

from django.conf import settings
from django.conf.urls import include, url

urlpatterns = [crudlfap.site.urlpattern]

if 'debug_toolbar' in settings.INSTALLED_APPS and settings.DEBUG:
    import debug_toolbar
    urlpatterns += [
        url(r'^__debug__/', include(debug_toolbar.urls)),
    ]



"""
**CRIMINALLY INVASIVE HACKS** in :py:class:`Factory`.
"""
import inspect


class FactoryMetaclass(type):
    """``__getattr__`` that ensures a first argument to getters.

    Makes the getter work both from class and instance

    Thanks to this, your `get_*()` methods will /maybe/ work in both
    cases::

        YourClass.foo   # calls get_foo(YourClass)
        YourClass().foo # calls get_foo(self)

    Don't code drunk.
    """

    def __getattr__(cls, attr):
        if attr.startswith('get_'):
            raise AttributeError('{} or {}'.format(attr[4:], attr))

        getter = getattr(cls, 'get_' + attr)

        if inspect.ismethod(getter):
            return getter()
        else:
            return getter(cls)

    def get_cls(cls):
        """Return the cls.

        did it go to far at this point ?
        """
        return cls


class Factory(metaclass=FactoryMetaclass):
    """Adds clumsy but automatic getter resolving.

    The `__getattr__` override makes this class try to call a get_*() method
    for variables that are not in `self.__dict__`.

    For example, when `self.foo` is evaluated `and 'foo' not in self.__dict__`
    then it will call the `self.get_foo()`

    If `self.get_foo()` returns None, it will try to get the result again from
    `self.__dict__`. Which means that we are going to witness this horrroorr::

        class YourEvil(Factory):
            def get_foo(self):
                self.calls += 1
                self.foo = 13

        assert YourEvil.foo == 13   # crime scene 1
        assert YourEvil.foo == 13   # crime scene 2
        assert YourEvil.calls == 1  # crime scene 3

    For the moment it is pretty clumsy because i tried to contain the
    criminality rate as low as possible meanwhile i like the work it does for
    me !
    """
    def __getattr__(self, attr):
        if attr.startswith('get_'):
            raise AttributeError('{} or {}()'.format(attr[4:], attr))

        getter = getattr(self, 'get_{}'.format(attr), None)

        if getter:
            methresult = getter()
            dictresult = self.__dict__.get(attr, None)
            if methresult is None and dictresult is not None:
                result = dictresult
            else:
                result = methresult
            return result

        # Try class methods
        return getattr(type(self), attr)

    @classmethod
    def clone(cls, *mixins, **attributes):
        """Return a subclass with the given attributes.

        If a model is found, it will prefix the class name with the model.
        """
        name = cls.__name__
        model = attributes.get('model', getattr(cls, 'model', None))
        if model is None:
            model = getattr(cls, 'model', None)
        if model and model.__name__ not in cls.__name__:
            name = '{}{}'.format(model.__name__, cls.__name__)
        return type(name, (cls,) + mixins, attributes)



# flake8: noqa: F401

from .crud import (
    CreateMixin,
    DeleteMixin,
    DetailMixin,
    HistoryMixin,
    ListMixin,
    UpdateMixin,
)
from .filter import FilterMixin
from .form import FormMixin
from .lock import LockMixin
from .menu import MenuMixin
from .model import ModelMixin
from .modelform import ModelFormMixin
from .object import ObjectMixin
from .objects import ObjectsMixin
from .objectform import ObjectFormMixin
from .objectsform import ObjectsFormMixin
from .search import SearchMixin
from .table import TableMixin
from .template import TemplateMixin
